from dataclasses import dataclass, field
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Union

from libcst import CSTNode

from cstvis.dto import Context
from cstvis.wrapper import CallableWrapper


@dataclass
class Collector:
    _filters: List[CallableWrapper[bool]] = field(default_factory=list)
    _converters: List[CallableWrapper[CSTNode]] = field(default_factory=list)

    def __add__(self, other: 'Collector') -> 'Collector':
        if not isinstance(other, type(self)):
            raise TypeError('Collector objects can only be added to other collector objects.')

        result = Collector()

        result._filters.extend(self._filters)
        result._filters.extend(other._filters)
        result._converters.extend(self._converters)
        result._converters.extend(other._converters)

        return result

    def filter(self, function: Optional[Union[Callable[[CSTNode], bool], Callable[[CSTNode, Context], bool]]] = None, meta: Optional[Dict[str, Any]] = None) -> Union[Union[Callable[[CSTNode], bool], Callable[[CSTNode, Context], bool]], Callable[[Union[Callable[[CSTNode], bool], Callable[[CSTNode, Context], bool]]], Union[Callable[[CSTNode], bool], Callable[[CSTNode, Context], bool]]]]:
        if function is None:
            return partial(self.filter, meta=meta)  # type: ignore[return-value]

        self._filters.append(CallableWrapper(function, meta=meta))
        return function

    def converter(self, function: Optional[Union[Callable[[CSTNode], CSTNode], Callable[[CSTNode, Context], CSTNode]]] = None, meta: Optional[Dict[str, Any]] = None) -> Union[Union[Callable[[CSTNode], CSTNode], Callable[[CSTNode, Context], CSTNode]], Callable[[Union[Callable[[CSTNode], CSTNode], Callable[[CSTNode, Context], CSTNode]]], Union[Callable[[CSTNode], CSTNode], Callable[[CSTNode, Context], CSTNode]]]]:
        if function is None:
            return partial(self.converter, meta=meta)  # type: ignore[return-value]

        self._converters.append(CallableWrapper(function, meta=meta))
        return function



import pytest
from full_match import match

from cstvis import Collector


def test_collections_in_different_collectors_are_not_same():
    first_collector = Collector()
    second_collector = Collector()

    assert isinstance(first_collector._filters, list)
    assert isinstance(first_collector._converters, list)

    assert len(first_collector._filters) == 0
    assert len(first_collector._converters) == 0

    assert first_collector._filters is not second_collector._filters
    assert first_collector._converters is not second_collector._converters


def test_collect_some_filter():
    collector = Collector()

    @collector.filter
    def some_filter(node, context):  # noqa: ARG001
        return True

    assert [x.function for x in collector._filters] == [some_filter]


def test_collect_some_converter():
    collector = Collector()

    @collector.converter
    def some_converter(node, context):  # noqa: ARG001
        return node

    assert [x.function for x in collector._converters] == [some_converter]


def test_add_two_collectors_with_converters():
    collector_1 = Collector()
    collector_2 = Collector()

    @collector_1.converter
    def some_converter_1(node, context):  # noqa: ARG001
        return node

    @collector_2.converter
    def some_converter_2(node, context):  # noqa: ARG001
        return node

    assert [x.function for x in collector_1._converters] == [some_converter_1]
    assert [x.function for x in collector_2._converters] == [some_converter_2]

    collector_3 = collector_1 + collector_2

    assert [x.function for x in collector_3._converters] == [some_converter_1, some_converter_2]


def test_add_two_collectors_with_filters():
    collector_1 = Collector()
    collector_2 = Collector()

    @collector_1.filter
    def some_converter_1(node, context):  # noqa: ARG001
        return node

    @collector_2.filter
    def some_converter_2(node, context):  # noqa: ARG001
        return node

    assert [x.function for x in collector_1._filters] == [some_converter_1]
    assert [x.function for x in collector_2._filters] == [some_converter_2]

    collector_3 = collector_1 + collector_2

    assert [x.function for x in collector_3._filters] == [some_converter_1, some_converter_2]


def test_add_wrong_things_to_collector():
    with pytest.raises(TypeError, match=match('Collector objects can only be added to other collector objects.')):
        Collector() + 1

    with pytest.raises(TypeError, match=match('Collector objects can only be added to other collector objects.')):
        Collector() + 'kek'



from collections import defaultdict
from functools import cached_property, partial
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Type,
    Union,
)

from libcst import CSTNode, metadata, parse_module

from cstvis.collector import Collector
from cstvis.dto import Context, Coordinate
from cstvis.transformers.super_transformer import SuperTransformer
from cstvis.visitors.bloodhound import Bloodhound
from cstvis.visitors.comments_aggregator import CommentsAggregator
from cstvis.wrapper import CallableWrapper


class Changer:
    def __init__(self, source: str, collector: Optional[Collector] = None) -> None:
        self.source = source
        self.module = parse_module(source)

        self.converters_by_types: Dict[Type[CSTNode], List[CallableWrapper[CSTNode]]] = defaultdict(list)
        self.filters_by_types: Dict[Type[CSTNode], List[CallableWrapper[bool]]] = defaultdict(list)

        if collector is not None:
            for collected_filter in collector._filters:
                self.filter(meta=collected_filter.meta)(collected_filter.function)  # type: ignore[call-arg, arg-type]
            for collected_converter in collector._converters:
                self.converter(meta=collected_converter.meta)(collected_converter.function)  # type: ignore[call-arg, arg-type]

    @cached_property
    def _comments_by_lines(self) -> Dict[int, str]:
        wrapper = metadata.MetadataWrapper(self.module)
        aggregator = CommentsAggregator()
        wrapper.visit(aggregator)
        return aggregator.comments

    def filter(self, function: Optional[Union[Callable[[CSTNode], bool], Callable[[CSTNode, Context], bool]]] = None, meta: Optional[Dict[str, Any]] = None) -> Union[Union[Callable[[CSTNode], bool], Callable[[CSTNode, Context], bool]], Callable[[Union[Callable[[CSTNode], bool], Callable[[CSTNode, Context], bool]]], Union[Callable[[CSTNode], bool], Callable[[CSTNode, Context], bool]]]]:
        if function is None:
            return partial(self.filter, meta=meta)  # type: ignore[return-value]

        wrapper = CallableWrapper(function, meta=meta)

        for annotation in wrapper.first_node_annotations:
            self.filters_by_types[annotation].append(wrapper)

        return function

    def converter(self, function: Optional[Union[Callable[[CSTNode], CSTNode], Callable[[CSTNode, Context], CSTNode]]] = None, meta: Optional[Dict[str, Any]] = None) -> Union[Union[Callable[[CSTNode], CSTNode], Callable[[CSTNode, Context], CSTNode]], Callable[[Union[Callable[[CSTNode], CSTNode], Callable[[CSTNode, Context], CSTNode]]], Union[Callable[[CSTNode], CSTNode], Callable[[CSTNode, Context], CSTNode]]]]:
        if function is None:
            return partial(self.converter, meta=meta)  # type: ignore[return-value]

        wrapper = CallableWrapper(function, meta=meta)

        for annotation in wrapper.first_node_annotations:
            self.converters_by_types[annotation].append(wrapper)

        return function

    def iterate_coordinates(self) -> Generator[Coordinate, None, None]:
        wrapper = metadata.MetadataWrapper(self.module)
        printer = Bloodhound(self.converters_by_types, self._comments_by_lines, self.filters_by_types)

        wrapper.visit(printer)
        yield from printer.coordinates

    def apply_coordinate(self, coordinate: Coordinate) -> str:
        wrapper = metadata.MetadataWrapper(self.module)
        modified = wrapper.visit(SuperTransformer(coordinate, self.converters_by_types, self._comments_by_lines, set()))
        return modified.code



# ruff: noqa: ARG001

from typing import Any, Union

import pytest
from full_match import match
from libcst import Add, CSTNode, Multiply, SimpleString, Subtract
from metacode import ParsedComment
from sigmatch import SignatureMismatchError

from cstvis import Changer, Collector, Context


@pytest.mark.parametrize(
    ['strings'],
    [
        ([
            'a = 5',
            'b = 12 * a #lol',
            'c = 12 + b # kek',
            'c *= 5',
            'c += 5',
            'a = c + 5',
        ],),
    ],
)
def test_just_iterate_add_coordinates(file, with_context, unfold):
    changer = Changer(file)

    if with_context:
        @unfold(changer.converter)
        def name_changer(node: Add, context: Context):
            return True
    else:
        @unfold(changer.converter)
        def name_changer(node: Add):
            return True

    coordinates = list(changer.iterate_coordinates())

    assert len(coordinates) == 2

    assert coordinates[0].file is None
    assert coordinates[0].class_name == 'Add'
    assert coordinates[0].start_line == 3
    assert coordinates[0].start_column == 7
    assert coordinates[0].start_line == 3
    assert coordinates[0].start_column == 7

    assert coordinates[1].file is None
    assert coordinates[1].class_name == 'Add'
    assert coordinates[1].start_line == 6
    assert coordinates[1].start_column == 6
    assert coordinates[1].start_line == 6
    assert coordinates[1].start_column == 6


@pytest.mark.parametrize(
    ['strings'],
    [
        ([
            'a = 5',
            'b = 12 * a #lol',
            'c = 12 + b # kek',
        ],),
    ],
)
def test_apply_one_change(file, with_context, unfold):
    changer = Changer(file)

    if with_context:
        @unfold(changer.converter)
        def change_add_to_sub(node: Add, context: Context):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )
    else:
        @unfold(changer.converter)
        def change_add_to_sub(node: Add):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

    results = []

    for coordinate in changer.iterate_coordinates():
        results.append(changer.apply_coordinate(coordinate))

    assert len(results) == 1
    assert results[0] == file.replace('+', '-')


@pytest.mark.parametrize(
    ['strings'],
    [
        ([
            'a = 5 + 6+ 7 +  8',
        ],),
    ],
)
def test_apply_two_changes_at_same_line(file, with_context, unfold):
    changer = Changer(file)

    if with_context:
        @unfold(changer.converter)
        def change_add_to_sub(node: Add, context: Context):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )
    else:
        @unfold(changer.converter)
        def change_add_to_sub(node: Add):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

    results = []

    for coordinate in changer.iterate_coordinates():
        results.append(changer.apply_coordinate(coordinate))

    assert len(results) == 3
    assert results == [
        'a = 5 - 6+ 7 +  8',
        'a = 5 + 6- 7 +  8',
        'a = 5 + 6+ 7 -  8',
    ]


@pytest.mark.parametrize(
    ['strings'],
    [
        ([
            'a = 5 + 6- 7',
        ],),
    ],
)
def test_to_different_changers_to_same_line(file, with_context, unfold):
    changer = Changer(file)

    if with_context:
        @unfold(changer.converter)
        def change_add_to_sub(node: Add, context: Context):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

        @unfold(changer.converter)
        def change_sub_to_add(node: Subtract, context: Context):
            return Add(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )
    else:
        @unfold(changer.converter)
        def change_add_to_sub(node: Add):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

        @unfold(changer.converter)
        def change_sub_to_add(node: Subtract):
            return Add(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

    results = []

    for coordinate in changer.iterate_coordinates():
        results.append(changer.apply_coordinate(coordinate))

    assert len(results) == 2
    assert results == [
        'a = 5 - 6- 7',
        'a = 5 + 6+ 7',
    ]


@pytest.mark.parametrize(
    ['strings'],
    [
        ([
            'a = 5 + 6- 7',
        ],),
    ],
)
def test_changing_function_with_wrong_number_of_parameters(file, unfold):
    changer = Changer(file)

    with pytest.raises(SignatureMismatchError, match=match('A function that takes a CST node and a context is expected.')):
        @unfold(changer.converter)
        def changing_function_1(node: Add, context: Context, something_else: str):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

    with pytest.raises(SignatureMismatchError, match=match('A function that takes a CST node and a context is expected.')):
        @unfold(changer.converter)
        def changing_function_2():
            ...


@pytest.mark.parametrize(
    ['strings', 'expected_comment'],
    [
        (['a = 5 + 6- 7'], None),
        (['a = 5 + 6- 7#'], ''),
        (['a = 5 + 6- 7# ololo!'], ' ololo!'),
        (['a = 5 + 6- 7# other_key: action'], ' other_key: action'),
        (['a = 5 + 6- 7#key: action'],  'key: action'),
        (['a = 5 + 6- 7# key: action'],  ' key: action'),
        (['a = 5 + 6- 7 # key: action'],  ' key: action'),
        (['a = 5 + 6- 7 # key: action# ololo!'],  ' key: action# ololo!'),
    ],
)
def test_read_comments(file, expected_comment, unfold):
    changer = Changer(file)

    comments_containers = []

    @unfold(changer.converter)
    def change_something(node: Add, context: Context):
        comments_containers.append(context.comment)
        return node

    for coordinate in changer.iterate_coordinates():
        changer.apply_coordinate(coordinate)

    assert comments_containers[0] == expected_comment


@pytest.mark.parametrize(
    ['strings', 'expected_metacodes'],
    [
        (['a = 5 + 6- 7'], []),
        (['a = 5 + 6- 7#'], []),
        (['a = 5 + 6- 7# ololo!'], []),
        (['a = 5 + 6- 7# other_key: action'], []),
        (['a = 5 + 6- 7# key: action'], [ParsedComment(key='key', command='action', arguments=[])]),
        (['a = 5 + 6- 7 # key: action'], [ParsedComment(key='key', command='action', arguments=[])]),
        (['a = 5 + 6- 7 # key: action# ololo!'], [ParsedComment(key='key', command='action', arguments=[])]),
    ],
)
def test_read_metacodes_from_comment(file, expected_metacodes, unfold):
    changer = Changer(file)

    metacodes_containers = []

    @unfold(changer.converter)
    def change_something(node: Add, context: Context):
        metacodes_containers.append(context.get_metacodes('key'))
        return node

    for coordinate in changer.iterate_coordinates():
        changer.apply_coordinate(coordinate)

    assert metacodes_containers[0] == expected_metacodes


@pytest.mark.parametrize(
    ['strings'],
    [
        ([
            'a = 5 + 6- 7',
        ],),
    ],
)
def test_filter_any_on(file, with_context, unfold):
    changer = Changer(file)

    if with_context:
        @unfold(changer.converter)
        def change_something(node: Add, context: Context):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

        @unfold(changer.filter)
        def filter_something(node: Any, context: Context) -> bool:
            return True

    else:
        @unfold(changer.converter)
        def change_something(node: Add):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

        @unfold(changer.filter)
        def filter_something(node: Any) -> bool:
            return True

    results = []

    for coordinate in changer.iterate_coordinates():
        results.append(changer.apply_coordinate(coordinate))

    assert len(results) == 1
    assert results[0] == file.replace('+', '-')


@pytest.mark.parametrize(
    ['strings'],
    [
        ([
            'a = 5 + 6- 7',
        ],),
    ],
)
def test_filter_any_off(file, with_context, unfold):
    changer = Changer(file)

    if with_context:
        @unfold(changer.converter)
        def change_something(node: Add, context: Context):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

        @unfold(changer.filter)
        def filter_something(node: Any, context: Context) -> bool:
            return False

    else:
        @unfold(changer.converter)
        def change_something(node: Add):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

        @unfold(changer.filter)
        def filter_something(node: Any) -> bool:
            return False

    results = []

    for coordinate in changer.iterate_coordinates():
        results.append(changer.apply_coordinate(coordinate))

    assert len(results) == 0


@pytest.mark.parametrize(
    ['strings'],
    [
        ([
            'a = 5 + 6- 7',
        ],),
    ],
)
def test_filter_cstnode_on(file, with_context, unfold):
    changer = Changer(file)

    if with_context:
        @unfold(changer.converter)
        def change_something(node: Add, context: Context):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

        @unfold(changer.filter)
        def filter_something(node: CSTNode, context: Context) -> bool:
            return True

    else:
        @unfold(changer.converter)
        def change_something(node: Add):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

        @unfold(changer.filter)
        def filter_something(node: CSTNode) -> bool:
            return True

    results = []

    for coordinate in changer.iterate_coordinates():
        results.append(changer.apply_coordinate(coordinate))

    assert len(results) == 1
    assert results[0] == file.replace('+', '-')


@pytest.mark.parametrize(
    ['strings'],
    [
        ([
            'a = 5 + 6- 7',
        ],),
    ],
)
def test_filter_cstnode_off(file, with_context, unfold):
    changer = Changer(file)

    if with_context:
        @unfold(changer.converter)
        def change_something(node: Add, context: Context):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

        @unfold(changer.filter)
        def filter_something(node: CSTNode, context: Context) -> bool:
            return False

    else:
        @unfold(changer.converter)
        def change_something(node: Add):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

        @unfold(changer.filter)
        def filter_something(node: CSTNode) -> bool:
            return False

    results = []

    for coordinate in changer.iterate_coordinates():
        results.append(changer.apply_coordinate(coordinate))

    assert len(results) == 0


@pytest.mark.parametrize(
    ['strings'],
    [
        ([
            'a = 5 + 6- 7',
        ],),
    ],
)
def test_filter_node_on(file, with_context, unfold):
    changer = Changer(file)

    if with_context:
        @unfold(changer.converter)
        def change_something(node: Add, context: Context):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

        @unfold(changer.filter)
        def filter_something(node: Add, context: Context) -> bool:
            return True

    else:
        @unfold(changer.converter)
        def change_something(node: Add):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

        @unfold(changer.filter)
        def filter_something(node: Add) -> bool:
            return True

    results = []

    for coordinate in changer.iterate_coordinates():
        results.append(changer.apply_coordinate(coordinate))

    assert len(results) == 1
    assert results[0] == file.replace('+', '-')


@pytest.mark.parametrize(
    ['strings'],
    [
        ([
            'a = 5 + 6- 7',
        ],),
    ],
)
def test_filter_node_off(file, with_context, unfold):
    changer = Changer(file)

    if with_context:
        @unfold(changer.converter)
        def change_something(node: Add, context: Context):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

        @unfold(changer.filter)
        def filter_something(node: Add, context: Context) -> bool:
            return False

    else:
        @unfold(changer.converter)
        def change_something(node: Add):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

        @unfold(changer.filter)
        def filter_something(node: Add) -> bool:
            return False

    results = []

    for coordinate in changer.iterate_coordinates():
        results.append(changer.apply_coordinate(coordinate))

    assert len(results) == 0


@pytest.mark.parametrize(
    ['strings'],
    [
        ([
            'a = 5 + 6- 7',
        ],),
    ],
)
def test_filter_other_node_on(file, with_context, unfold):
    changer = Changer(file)

    if with_context:
        @unfold(changer.converter)
        def change_something(node: Add, context: Context):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

        @unfold(changer.filter)
        def filter_something(node: Subtract, context: Context) -> bool:
            return True

    else:
        @unfold(changer.converter)
        def change_something(node: Add):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

        @unfold(changer.filter)
        def filter_something(node: Subtract) -> bool:
            return True

    results = []

    for coordinate in changer.iterate_coordinates():
        results.append(changer.apply_coordinate(coordinate))

    assert len(results) == 1
    assert results[0] == file.replace('+', '-')


@pytest.mark.parametrize(
    ['strings'],
    [
        ([
            'a = 5 + 6- 7',
        ],),
    ],
)
def test_filter_other_node_off(file, with_context, unfold):
    changer = Changer(file)

    if with_context:
        @unfold(changer.converter)
        def change_something(node: Add, context: Context):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

        @unfold(changer.filter)
        def filter_something(node: Subtract, context: Context) -> bool:
            return False

    else:
        @unfold(changer.converter)
        def change_something(node: Add):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

        @unfold(changer.filter)
        def filter_something(node: Subtract) -> bool:
            return False

    results = []

    for coordinate in changer.iterate_coordinates():
        results.append(changer.apply_coordinate(coordinate))

    assert len(results) == 1
    assert results[0] == file.replace('+', '-')


def test_converter_with_no_annotation(with_context, unfold):
    changer = Changer('1')

    if with_context:
        @unfold(changer.converter)
        def converter_func(node, context):
            return node
    else:
        @unfold(changer.converter)
        def converter_func(node):
            return node

    assert [changer.apply_coordinate(coordinate) for coordinate in changer.iterate_coordinates()]


def test_converter_with_any_annotation(with_context, unfold):
    changer = Changer('1')

    if with_context:
        @unfold(changer.converter)
        def converter_func(node: Any, context):
            return node
    else:
        @unfold(changer.converter)
        def converter_func(node: Any):
            return node

    assert [changer.apply_coordinate(coordinate) for coordinate in changer.iterate_coordinates()]


def test_converter_with_cstnode_annotation_restriction(with_context, unfold):
    changer = Changer('1')

    if with_context:
        @unfold(changer.converter)
        def converter_func(node: CSTNode, context):
            return node
    else:
        @unfold(changer.converter)
        def converter_func(node: CSTNode):
            return node

    assert [changer.apply_coordinate(coordinate) for coordinate in changer.iterate_coordinates()]


def test_convert_str(with_context, unfold):
    changer = Changer('a = "kek"')

    nodes = []

    if with_context:
        @unfold(changer.converter)
        def converter_func(node: str, context: Context):
            nodes.append(node)
            return node
    else:
        @unfold(changer.converter)
        def converter_func(node: str):
            nodes.append(node)
            return node

    for coordinate in changer.iterate_coordinates():
        changer.apply_coordinate(coordinate)

    assert len(nodes) == 1
    assert isinstance(nodes[0], SimpleString)


def test_convert_float(with_context, unfold):
    changer = Changer('a = 5.0')

    if with_context:
        @unfold(changer.converter)
        def converter_func(node: float, context: Context):
            return node.with_changes(value=repr(node.evaluated_value + 1))  # type: ignore[attr-defined]
    else:
        @unfold(changer.converter)
        def converter_func(node: float):
            return node.with_changes(value=repr(node.evaluated_value + 1))  # type: ignore[attr-defined]

    for coordinate in changer.iterate_coordinates():
        changer.apply_coordinate(coordinate)

    assert [changer.apply_coordinate(coordinate) for coordinate in changer.iterate_coordinates()] == ['a = 6.0']


def test_filter_with_wrong_number_of_parameters(unfold):
    changer = Changer('a = 5')

    with pytest.raises(SignatureMismatchError, match=match('A function that takes a CST node and a context is expected.')):
        @unfold(changer.filter)
        def filter_func(node: Add, context: Context, extra_param: str):
            return True

    with pytest.raises(SignatureMismatchError, match=match('A function that takes a CST node and a context is expected.')):
        @unfold(changer.filter)
        def filter_func():
            return True


def test_filter_with_invalid_annotation(unfold):
    changer = Changer('a = 5')

    class SomeClass:
        pass

    with pytest.raises(TypeError, match=match('The type annotation for the first argument of the function must be descended from the libcst.CSTNode class.')):
        @unfold(changer.filter)
        def filter_func(node: SomeClass, context: Context):
            return True


def test_two_converters_for_same_node(with_context, unfold):
    changer = Changer('5 + 5')

    if with_context:
        @unfold(changer.converter)
        def converter1(node: Add, context: Context):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

        @unfold(changer.converter)
        def converter2(node: Add, context: Context):
            return Multiply(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

    else:
        @unfold(changer.converter)
        def converter1(node: Add):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

        @unfold(changer.converter)
        def converter2(node: Add):
            return Multiply(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

    assert set(changer.apply_coordinate(coordinate) for coordinate in changer.iterate_coordinates()) == {'5 - 5', '5 * 5'}


def test_use_collector_for_converter(with_context, unfold):
    collector = Collector()

    if with_context:
        @unfold(collector.converter)
        def some_converter(node: Add, context: Context):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )
    else:
        @unfold(collector.converter)
        def some_converter(node: Add):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

    changer = Changer('a = 5 + 5', collector=collector)

    assert [changer.apply_coordinate(coordinate) for coordinate in changer.iterate_coordinates()] == ['a = 5 - 5']


def test_use_collector_for_converter_and_filter(with_context, unfold):
    collector = Collector()

    filters_value = False

    if with_context:
        @unfold(collector.converter)
        def some_converter(node: Add, context: Context):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

        @unfold(collector.filter)
        def some_filter(node: Add, context: Context):
            return filters_value

    else:
        @unfold(collector.converter)
        def some_converter(node: Add):
            return Subtract(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

        @unfold(collector.filter)
        def some_filter(node: Add):
            return filters_value

    changer = Changer('a = 5 + 5', collector=collector)

    assert [changer.apply_coordinate(coordinate) for coordinate in changer.iterate_coordinates()] == []

    filters_value = True

    assert [changer.apply_coordinate(coordinate) for coordinate in changer.iterate_coordinates()] == ['a = 5 - 5']


def test_union_with_csts(with_context, unfold):
    changer = Changer('5 - 5 + 5')

    if with_context:
        @unfold(changer.converter)
        def some_converter(node: Union[Add, Subtract], context: Context):
            return Multiply(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )
    else:
        @unfold(changer.converter)
        def some_converter(node: Union[Add, Subtract]):
            return Multiply(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

    assert [changer.apply_coordinate(coordinate) for coordinate in changer.iterate_coordinates()] == ['5 * 5 + 5', '5 - 5 * 5']


def test_union_with_union_with_csts(with_context, unfold):
    changer = Changer('5 - 5 + 5')

    if with_context:
        @unfold(changer.converter)
        def some_converter(node: Union[Add, Union[Multiply, Subtract]], context: Context):
            return Multiply(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )
    else:
        @unfold(changer.converter)
        def some_converter(node: Union[Add, Union[Multiply, Subtract]]):
            return Multiply(
                whitespace_before=node.whitespace_before,
                whitespace_after=node.whitespace_after,
            )

    assert set(changer.apply_coordinate(coordinate) for coordinate in changer.iterate_coordinates()) == {'5 * 5 + 5', '5 - 5 * 5'}


def test_convert_plus_one(with_context, unfold):
    changer = Changer('5 - 5 + 5')

    if with_context:
        @unfold(changer.converter)
        def convert_ints(node: int, context):
            return node.with_changes(value=repr(node.evaluated_value + 1))  # type: ignore[attr-defined]
    else:
        @unfold(changer.converter)
        def convert_ints(node: int):
            return node.with_changes(value=repr(node.evaluated_value + 1))  # type: ignore[attr-defined]

    assert set(changer.apply_coordinate(coordinate) for coordinate in changer.iterate_coordinates()) == {'6 - 5 + 5', '5 - 6 + 5', '5 - 5 + 6'}


def test_converter_for_any(with_context, unfold):
    changer = Changer('5 - 5 + 5')

    nodes = []

    if with_context:
        @unfold(changer.converter)
        def do_something(node: Any, context):
            nodes.append(nodes)
            return node
    else:
        @unfold(changer.converter)
        def do_something(node: Any):
            nodes.append(nodes)
            return node

    [changer.apply_coordinate(coordinate) for coordinate in changer.iterate_coordinates()]
    assert len(nodes) > 10


def test_if_node_is_not_exist_nothing_changed(with_context, unfold):
    changer = Changer('5 - 5 + 5')

    if with_context:
        @unfold(changer.converter)
        def do_something(node: float, context):
            return node
    else:
        @unfold(changer.converter)
        def do_something(node: float):
            return node

    assert [changer.apply_coordinate(coordinate) for coordinate in changer.iterate_coordinates()] == []


def test_get_function_id_from_itself(unfold):
    changer = Changer('5 - 5 + 5')

    @unfold(changer.converter)
    def do_something(node: float, context):
        return node

    @unfold(changer.filter)
    def filter_something(node: float, context):
        return False

    converter = list(changer.converters_by_types.values())[0][0]  # noqa: RUF015
    filter = list(changer.filters_by_types.values())[0][0]  # noqa: RUF015, A001

    assert converter.get_function_id() == 'tests.test_changer:do_something:921'
    assert filter.get_function_id() == 'tests.test_changer:filter_something:925'


def test_wrong_converter_and_wrong_filter(unfold):
    changer = Changer('5 - 5 + 5')

    with pytest.raises(SignatureMismatchError, match=match('A function that takes a CST node and a context is expected.')):
        @unfold(changer.converter)
        def do_something_1():
            ...

    with pytest.raises(SignatureMismatchError, match=match('A function that takes a CST node and a context is expected.')):
        @unfold(changer.converter)
        def do_something_2(a, b, c):
            ...

    with pytest.raises(SignatureMismatchError, match=match('A function that takes a CST node and a context is expected.')):
        @unfold(changer.filter)
        def filter_something_1():
            return False

    with pytest.raises(SignatureMismatchError, match=match('A function that takes a CST node and a context is expected.')):
        @unfold(changer.filter)
        def filter_something_2(a, b, c):
            return False


def test_pass_meta_dict_to_converter():
    bread_crumbs = []
    changer = Changer('5 - 5 + 5')
    meta = {'key': 123}

    @changer.converter(meta=meta)
    def some_converter(node: Add, context: Context):
        bread_crumbs.append(context.meta)
        return Multiply(
            whitespace_before=node.whitespace_before,
            whitespace_after=node.whitespace_after,
        )

    [changer.apply_coordinate(coordinate) for coordinate in changer.iterate_coordinates()]

    assert bread_crumbs == [meta]
    assert bread_crumbs[0] is not meta


def test_pass_meta_dict_to_filter(unfold):
    bread_crumbs = []
    changer = Changer('5 - 5 + 5')
    meta = {'key': 123}

    @unfold(changer.converter)
    def some_converter(node: Add):
        return Multiply(
            whitespace_before=node.whitespace_before,
            whitespace_after=node.whitespace_after,
        )

    @changer.filter(meta=meta)
    def filter_something(node: Add, context: Context):
        bread_crumbs.append(context.meta)
        return False

    [changer.apply_coordinate(coordinate) for coordinate in changer.iterate_coordinates()]

    assert bread_crumbs == [meta]
    assert bread_crumbs[0] is not meta


def test_pass_meta_dict_to_converter_throw_collector():
    bread_crumbs = []
    collector = Collector()
    meta = {'key': 123}

    @collector.converter(meta=meta)
    def some_converter(node: Add, context: Context):
        bread_crumbs.append(context.meta)
        return Multiply(
            whitespace_before=node.whitespace_before,
            whitespace_after=node.whitespace_after,
        )

    changer = Changer('5 - 5 + 5', collector=collector)
    [changer.apply_coordinate(coordinate) for coordinate in changer.iterate_coordinates()]

    assert bread_crumbs == [meta]
    assert bread_crumbs[0] is not meta


def test_pass_meta_dict_to_filter_throw_collector(unfold):
    bread_crumbs = []
    collector = Collector()
    meta = {'key': 123}

    @unfold(collector.converter)
    def some_converter(node: Add):
        return Multiply(
            whitespace_before=node.whitespace_before,
            whitespace_after=node.whitespace_after,
        )

    @collector.filter(meta=meta)
    def filter_something(node: Add, context: Context):
        bread_crumbs.append(context.meta)
        return False

    changer = Changer('5 - 5 + 5', collector=collector)
    [changer.apply_coordinate(coordinate) for coordinate in changer.iterate_coordinates()]

    assert bread_crumbs == [meta]
    assert bread_crumbs[0] is not meta


def test_it_passes_2_arguments_if_possible(unfold):
    changer = Changer('5 + 4')

    contexts = []

    @unfold(changer.converter)
    def change_something(node: Add, context=None):
        contexts.append(context)
        return node

    for coordinate in changer.iterate_coordinates():
        changer.apply_coordinate(coordinate)

    assert len(contexts) == 1
    assert isinstance(contexts[0], Context)



from typing import Dict

from libcst import Comment, CSTNode, CSTVisitor, metadata


class CommentsAggregator(CSTVisitor):
    METADATA_DEPENDENCIES = (metadata.PositionProvider,)

    def __init__(self) -> None:
        self.comments: Dict[int, str] = {}

    def on_visit(self, node: CSTNode) -> bool:
        if isinstance(node, Comment):
            position = self.get_metadata(metadata.PositionProvider, node)
            self.comments[position.start.line] = node.value[1:]
        return True



from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
)

# TODO: Delete this try-except if Python's version is >= 3.10
try:
    from types import UnionType  # type: ignore[attr-defined, unused-ignore]
except ImportError:  # pragma: no cover
    from typing import Union as UnionType  # type: ignore[assignment, unused-ignore]

from functools import cached_property
from inspect import _empty, isclass, signature

from libcst import CSTNode, Float, Integer, SimpleString
from printo import repred
from sigmatch import PossibleCallMatcher, SignatureMismatchError

from cstvis.dto import Context

FilterOrConverterReturnValue = TypeVar('FilterOrConverterReturnValue')

@repred(prefer_positional=True)  # type: ignore[call-overload]
class CallableWrapper(Generic[FilterOrConverterReturnValue]):
    matcher = PossibleCallMatcher('.') + PossibleCallMatcher('..')

    def __init__(self, function: Union[Callable[[CSTNode], FilterOrConverterReturnValue], Callable[[CSTNode, Context], FilterOrConverterReturnValue]], meta: Optional[Dict[str, Any]] = None) -> None:
        if not self.matcher.match(function):
            raise SignatureMismatchError('A function that takes a CST node and a context is expected.')

        self.function = function
        self.meta = meta

    def __call__(self, node: CSTNode, context: Context) -> FilterOrConverterReturnValue:
        if PossibleCallMatcher('..').match(self.function):
            context.meta = self.meta.copy() if isinstance(self.meta, dict) else None
            return self.function(node, context)  # type: ignore[call-arg]

        return self.function(node)  # type: ignore[call-arg]

    def get_function_id(self) -> str:
        return f'{self.function.__module__}:{self.function.__name__}:{self.function.__code__.co_firstlineno}'

    @cached_property
    def first_node_annotations(self) -> List[Type[CSTNode]]:
        converter_signature = signature(self.function)

        first_parameter = converter_signature.parameters[next(iter(converter_signature.parameters))]
        super_annotation = first_parameter.annotation if first_parameter.annotation is not _empty and first_parameter.annotation is not Any else CSTNode

        return self._separate_annotation(super_annotation)

    def _separate_annotation(self, annotation: Union[Type[CSTNode], Any]) -> List[Type[CSTNode]]:
        if isclass(annotation) and issubclass(annotation, CSTNode):
            return [annotation]

        if get_origin(annotation) is Union or get_origin(annotation) is UnionType:
            result = []
            for argument in get_args(annotation):
                result += self._separate_annotation(argument)
            return result

        if annotation is int:
            return [Integer]

        if annotation is float:
            return [Float]

        if annotation is str:
            return [SimpleString]

        raise TypeError('The type annotation for the first argument of the function must be descended from the libcst.CSTNode class.')



from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from metacode import ParsedComment, parse


@dataclass
class Coordinate:
    file: Optional[Path]
    class_name: str
    start_line: int
    start_column: int
    end_line: int
    end_column: int
    converter_id: Optional[str] = None

@dataclass
class Context:
    coordinate: Coordinate
    comment: Optional[str]
    meta: Optional[Dict[str, Any]] = None

    def get_metacodes(self, key: Union[str, List[str]]) -> List[ParsedComment]:
        if self.comment is None:
            return []
        return parse(self.comment, key)



from typing import Any, Callable, Dict, List, Set, Type

import libcst.matchers as matchers_module
from libcst import CSTNode, metadata
from libcst.matchers import (
    BaseMatcherNode,
    MatcherDecoratableTransformer,
    TypeOf,
    leave,
)

from cstvis.dto import Context, Coordinate
from cstvis.wrapper import CallableWrapper


def get_all_matcher_nodes() -> List[BaseMatcherNode]:
    result = []

    for name in dir(matchers_module):
        attribute = getattr(matchers_module, name)
        try:
            if issubclass(attribute, BaseMatcherNode) and attribute is not BaseMatcherNode and attribute is not TypeOf:
                result.append(attribute())
        except TypeError:
            pass

    return result

def leave_all(function: Callable[[Any, CSTNode, CSTNode], CSTNode]) -> Callable[[Any, CSTNode, CSTNode], CSTNode]:
    for matcher in get_all_matcher_nodes():
        function = leave(matcher)(function)

    return function


class SuperTransformer(MatcherDecoratableTransformer):
    METADATA_DEPENDENCIES = (metadata.PositionProvider,)

    def __init__(
        self,
        target_coordinate: Coordinate,
        nodes_mapping: Dict[Type[CSTNode], List[CallableWrapper[CSTNode]]],
        comments: Dict[int, str],
        nodes_ids: Set[int],
    ):
        self.target_coordinate = target_coordinate
        self.nodes_mapping = nodes_mapping
        self.comments = comments
        self.nodes_ids = nodes_ids

        super().__init__()

    @leave_all
    def leave(self, original_node, updated_node):  # type: ignore[no-untyped-def]
        if id(original_node) in self.nodes_ids:
            return updated_node
        self.nodes_ids.add(id(original_node))

        position = self.get_metadata(metadata.PositionProvider, original_node)
        coordinate = Coordinate(
            file=None,
            class_name=original_node.__class__.__name__,
            start_line=position.start.line,
            start_column=position.start.column,
            end_line=position.end.line,
            end_column=position.end.column,
        )
        target_coordinate_without_converter_id = Coordinate(
            file=None,
            class_name=self.target_coordinate.class_name,
            start_line=self.target_coordinate.start_line,
            start_column=self.target_coordinate.start_column,
            end_line=self.target_coordinate.end_line,
            end_column=self.target_coordinate.end_column,
        )

        converters = self.nodes_mapping.get(type(original_node), []) + self.nodes_mapping.get(CSTNode, [])  # type: ignore[type-abstract]

        if coordinate == target_coordinate_without_converter_id and converters:
            context = Context(coordinate, self.comments.get(coordinate.start_line))
            for converter in converters:  # pragma: no branch
                if converter.get_function_id() == self.target_coordinate.converter_id:
                    return converter(updated_node, context)
        return updated_node



from copy import deepcopy
from typing import Dict, List, Type

from libcst import CSTNode, CSTVisitor, metadata

from cstvis.dto import Context, Coordinate
from cstvis.wrapper import CallableWrapper


class Bloodhound(CSTVisitor):
    METADATA_DEPENDENCIES = (metadata.PositionProvider,)

    def __init__(
        self,
        nodes_mapping: Dict[Type[CSTNode], List[CallableWrapper[CSTNode]]],
        comments: Dict[int, str],
        filters: Dict[Type[CSTNode], List[CallableWrapper[bool]]],
    ) -> None:
        self.coordinates: List[Coordinate] = []
        self.nodes_mapping = nodes_mapping
        self.comments = comments
        self.filters = filters

    def on_visit(self, node: CSTNode) -> bool:
        position = self.get_metadata(metadata.PositionProvider, node)
        coordinate = Coordinate(
            file=None,
            class_name=node.__class__.__name__,
            start_line=position.start.line,
            start_column=position.start.column,
            end_line=position.end.line,
            end_column=position.end.column,
        )

        converters = self.nodes_mapping.get(type(node), []) + self.nodes_mapping.get(CSTNode, [])  # type: ignore[type-abstract]

        if converters:
            filters = self.filters.get(type(node), []) + self.filters.get(CSTNode, [])  # type: ignore[type-abstract]
            context = Context(coordinate, self.comments.get(coordinate.start_line))
            if filters:
                for filter_function in filters:
                    if not filter_function(node, context):
                        return True
            for converter_id in set([x.get_function_id() for x in converters]):
                emitting_coordinate = deepcopy(coordinate)
                emitting_coordinate.converter_id = converter_id
                self.coordinates.append(
                    emitting_coordinate,
                )

        return True
